using Humanizer;
using ShellProgressBar;
using System.Text.Json;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Stateless.Methods;
using WindowsShortcutFactory;

namespace View_by_Distance.Map.Models;

public class MapLogic
{

    protected readonly Dictionary<int, List<int>> _SkipCollection;
    protected readonly List<PersonContainer> _NotMappedPersonContainers;
    protected readonly Dictionary<long, PersonContainer> _PersonKeyToPersonContainer;
    protected readonly Dictionary<long, (long LCL, long Minimum, long Maximum, long UCL)> _PersonKeyToRanges;
    protected readonly Dictionary<int, Dictionary<int, PersonContainer[]>> _IdThenNormalizedPixelPercentageToPersonContainers;

    public Dictionary<int, int[]> KeyValuePairs => throw new NotImplementedException();
    public Dictionary<int, int[]> IndicesFromNew => throw new NotImplementedException();

    private readonly long _Ticks;
    private readonly Serilog.ILogger? _Log;
    private readonly int _MaxDegreeOfParallelism;
    private readonly Configuration? _Configuration;
    private readonly string _EDistanceContentTicksDirectory;

    public MapLogic(int maxDegreeOfParallelism, Shared.Models.Properties.IPropertyConfiguration propertyConfiguration, Configuration? configuration, long ticks, PersonContainer[] personContainers, string a2PeopleSingletonDirectory, string eDistanceContentDirectory, Mapping[] mappingCollection, Shared.Models.Methods.IMapLogicSupport? faceDistance)
    {
        _Ticks = ticks;
        _Configuration = configuration;
        _Log = Serilog.Log.ForContext<MapLogic>();
        _MaxDegreeOfParallelism = maxDegreeOfParallelism;
        if (_Log is null)
        { }
        if (propertyConfiguration.VerifyToSeason is null || !propertyConfiguration.VerifyToSeason.Any())
            throw new Exception();
        if (_MaxDegreeOfParallelism == 0)
        { }
        string json;
        string fullPath;
        List<KeyValuePair<int, int[]>>? collection;
        Dictionary<int, List<int>> skipCollection = new();
        List<PersonContainer> notMappedPersonContainers = new();
        Dictionary<long, PersonContainer> personKeyToPersonContainer = new();
        string? rootDirectoryParent = Path.GetDirectoryName(propertyConfiguration.RootDirectory);
        Dictionary<long, (long LCL, long Minimum, long Maximum, long UCL)> personKeyToRanges = new();
        string eDistanceContentTicksDirectory = Path.Combine(eDistanceContentDirectory, $"({ticks})");
        Dictionary<int, Dictionary<int, PersonContainer[]>> idThenNormalizedPixelPercentageToPersonContainers = new();
        for (int i = 1; i < 5; i++)
            _ = IPath.DeleteEmptyDirectories(eDistanceContentDirectory);
        _ = IPath.DeleteEmptyDirectories(eDistanceContentDirectory);
        if (string.IsNullOrEmpty(rootDirectoryParent))
            throw new NullReferenceException(nameof(rootDirectoryParent));
        if (!Directory.Exists(eDistanceContentDirectory))
            _ = Directory.CreateDirectory(eDistanceContentDirectory);
        if (!Directory.Exists(eDistanceContentTicksDirectory))
            _ = Directory.CreateDirectory(eDistanceContentTicksDirectory);
        if (configuration is not null && faceDistance is not null)
        {
            List<PersonContainer> personContainerCollection = new(personContainers);
            Stateless.MapLogic.Set(propertyConfiguration,
                                   configuration,
                                   ticks,
                                   personContainerCollection,
                                   a2PeopleSingletonDirectory,
                                   eDistanceContentDirectory,
                                   mappingCollection,
                                   faceDistance,
                                   personKeyToPersonContainer,
                                   personKeyToRanges,
                                   notMappedPersonContainers,
                                   skipCollection,
                                   idThenNormalizedPixelPercentageToPersonContainers);
            if (personContainerCollection.Count == personContainers.Length)
                throw new NotSupportedException();
        }
        foreach (string propertyContentCollectionFile in propertyConfiguration.PropertyContentCollectionFiles)
        {
            fullPath = Path.GetFullPath(string.Concat(rootDirectoryParent, propertyContentCollectionFile));
            if (fullPath.Contains(propertyConfiguration.RootDirectory))
                continue;
            if (!File.Exists(fullPath))
                continue;
            json = File.ReadAllText(fullPath);
            collection = JsonSerializer.Deserialize<List<KeyValuePair<int, int[]>>>(json);
            if (collection is null)
                throw new NullReferenceException(nameof(collection));
        }
        _SkipCollection = skipCollection;
        _PersonKeyToRanges = personKeyToRanges;
        _NotMappedPersonContainers = notMappedPersonContainers;
        _PersonKeyToPersonContainer = personKeyToPersonContainer;
        _EDistanceContentTicksDirectory = eDistanceContentTicksDirectory;
        _IdThenNormalizedPixelPercentageToPersonContainers = idThenNormalizedPixelPercentageToPersonContainers;
    }

    public MapLogic(int maxDegreeOfParallelism, Shared.Models.Properties.IPropertyConfiguration propertyConfiguration, Configuration? configuration, long ticks, PersonContainer[] personContainers, string a2PeopleSingletonDirectory, string eDistanceContentDirectory) :
        this(maxDegreeOfParallelism, propertyConfiguration, configuration, ticks, personContainers, a2PeopleSingletonDirectory, eDistanceContentDirectory, Array.Empty<Mapping>(), null)
    { }

    public override string ToString()
    {
        string result = JsonSerializer.Serialize(this, new JsonSerializerOptions() { WriteIndented = true });
        return result;
    }

    public int AddToMapping(Mapping[] mappingCollection)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        int result = 0;
        const int zero = 0;
        string mappingSegmentB;
        PersonBirthday personBirthday;
        PersonContainer[]? collection;
        int by = Stateless.IMapLogic.Mapping;
        List<PersonContainer> personContainers = new();
        Dictionary<int, PersonContainer[]>? keyValuePairs;
        foreach (Mapping mapping in mappingCollection)
        {
            personContainers.Clear();
            if (!_IdThenNormalizedPixelPercentageToPersonContainers.TryGetValue(mapping.MappingFromItem.Id, out keyValuePairs))
                result += 1;
            else
            {
                if (!keyValuePairs.TryGetValue(mapping.MappingFromLocation.NormalizedPixelPercentage, out collection))
                    result += 1;
                else
                    personContainers.AddRange(collection);
            }
            foreach (PersonContainer personContainer in personContainers)
            {
                if (personContainer.Key is null || personContainer.Birthdays is null || !personContainer.Birthdays.Any())
                    continue;
                personBirthday = personContainer.Birthdays[zero];
                mappingSegmentB = Stateless.MapLogic.GetMappingSegmentB(_Ticks, personBirthday, personContainer.ApproximateYears, mapping.MappingFromItem);
                mapping.UpdateMappingFromPerson(personContainer.ApproximateYears, by, personContainer.DisplayDirectoryName, personBirthday, mappingSegmentB);
            }
        }
        return result;
    }

    private void SaveContainers(int totalNotMapped, int? updated, List<SaveContainer> saveContainers)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        string checkFile;
        string sourceFile;
        WindowsShortcut windowsShortcut;
        string[] directories = (from l in saveContainers select l.Directory).Distinct().ToArray();
        int totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - _Ticks).TotalSeconds);
        string message;
        if (updated is null)
            message = $") {saveContainers.Count:000} save(s) - {totalNotMapped} Total not Mapped - {totalSeconds} total second(s)";
        else
            message = $") {saveContainers.Count:000} save(s) - {totalNotMapped} Total not Mapped - {updated} Updated - {totalSeconds} total second(s)";
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        foreach (string directory in directories)
        {
            if (string.IsNullOrEmpty(directory))
                continue;
            if (!Directory.Exists(directory))
                _ = Directory.CreateDirectory(directory);
        }
        using ProgressBar progressBar = new(saveContainers.Count, message, options);
        foreach (SaveContainer saveContainer in saveContainers)
        {
            progressBar.Tick();
            if (string.IsNullOrEmpty(saveContainer.Directory) || string.IsNullOrEmpty(saveContainer.CheckFile) || saveContainer.FaceFileHolder is null)
                continue;
            if (!saveContainer.FaceFileHolder.Exists && saveContainer.ResizedFileHolder is not null && saveContainer.ResizedFileHolder.Exists)
            {
                checkFile = saveContainer.CheckFile;
                sourceFile = saveContainer.ResizedFileHolder.FullName;
            }
            else if (saveContainer.FaceFileHolder.Exists)
            {
                sourceFile = saveContainer.FaceFileHolder.FullName;
                checkFile = $"{saveContainer.CheckFile}{_Configuration.FacesFileNameExtension}";
            }
            else
                continue;
            if (File.Exists(checkFile))
                continue;
            File.Copy(sourceFile, checkFile);
            if (saveContainer.HiddenFaceFileHolder is not null && saveContainer.HiddenFaceFileHolder.Exists)
            {
                sourceFile = saveContainer.HiddenFaceFileHolder.FullName;
                checkFile = $"{saveContainer.CheckFile}{_Configuration.FacesHiddenFileNameExtension}";
            }
            else if (saveContainer.FacePartsFileHolder is not null && saveContainer.FacePartsFileHolder.Exists)
            {
                sourceFile = saveContainer.FacePartsFileHolder.FullName;
                checkFile = $"{saveContainer.CheckFile}{_Configuration.FacePartsFileNameExtension}";
            }
            if (File.Exists(checkFile))
                continue;
            File.Copy(sourceFile, checkFile);
        }
        foreach (SaveContainer saveContainer in saveContainers)
        {
            if (string.IsNullOrEmpty(saveContainer.ShortcutFile) || saveContainer.ResizedFileHolder is null || !saveContainer.ResizedFileHolder.Exists)
                continue;
            try
            {
                windowsShortcut = new() { Path = saveContainer.ResizedFileHolder.FullName };
                windowsShortcut.Save(saveContainer.ShortcutFile);
                windowsShortcut.Dispose();
            }
            catch (Exception)
            { }
        }
    }

    public void SaveMapped(string dFacesContentDirectory, string d2FacePartsContentDirectory, Mapping[] mappingCollection, Dictionary<int, Dictionary<int, Mapping>> idToNormalizedPixelPercentageToMapping, int totalNotMapped)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        int? updated = null;
        int? useFiltersCounter = null;
        string mappingDirectory = Path.Combine(_EDistanceContentTicksDirectory, nameof(Stateless.IMapLogic.Mapping));
        List<SaveContainer> saveContainers = Stateless.MapLogic.GetSaveContainers(_Configuration, _EDistanceContentTicksDirectory, dFacesContentDirectory, d2FacePartsContentDirectory, mappingCollection, idToNormalizedPixelPercentageToMapping, useFiltersCounter, saveMapped: true);
        SaveContainers(totalNotMapped, updated, saveContainers);
        if (!string.IsNullOrEmpty(_EDistanceContentTicksDirectory) && Directory.Exists(mappingDirectory))
            Stateless.MapLogic.SaveMappingShortcuts(mappingDirectory);
    }

    private List<(long, long, long, long)> GetPersonKeysRangesCollection(PersonContainer[] personContainers)
    {
        List<(long, long, long, long)> results = new();
        (long, long, long, long) singleton;
        foreach (PersonContainer personContainer in personContainers)
        {
            if (personContainer.Key is null)
                continue;
            if (!_PersonKeyToRanges.TryGetValue(personContainer.Key.Value, out singleton))
                continue;
            results.Add(singleton);
        }
        return results;
    }

    public bool Used(FaceDistance faceDistanceEncoding)
    {
        bool result = false;
        if (faceDistanceEncoding.NormalizedPixelPercentage is null)
            throw new NotSupportedException();
        List<int>? normalizedPixelPercentages;
        Dictionary<int, PersonContainer[]>? normalizedPixelPercentageToPersonContainers;
        if (_SkipCollection.TryGetValue(faceDistanceEncoding.Id, out normalizedPixelPercentages))
        {
            if (normalizedPixelPercentages.Contains(faceDistanceEncoding.NormalizedPixelPercentage.Value))
                result = true;
        }
        if (!result && _IdThenNormalizedPixelPercentageToPersonContainers.TryGetValue(faceDistanceEncoding.Id, out normalizedPixelPercentageToPersonContainers))
        {
            if (normalizedPixelPercentageToPersonContainers.ContainsKey(faceDistanceEncoding.NormalizedPixelPercentage.Value))
                result = true;
        }
        return result;
    }

    public List<Sorting> GetSortingCollection(int i, FaceDistance faceDistanceEncoding, List<FaceDistance> faceDistanceLengths)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        List<Sorting> results = new();
        Sorting sorting;
        FaceDistance faceDistanceLength;
        List<int>? normalizedPixelPercentages;
        Dictionary<int, PersonContainer[]>? normalizedPixelPercentageToPersonContainers;
        List<(long lcl, long minimum, long maximum, long ucl)> personKeysRangesCollection;
        for (int j = 0; j < faceDistanceLengths.Count; j++)
        {
            if (faceDistanceEncoding.NormalizedPixelPercentage is null)
                throw new NotSupportedException();
            if (j == i)
                continue;
            if (_SkipCollection.TryGetValue(faceDistanceEncoding.Id, out normalizedPixelPercentages))
            {
                if (normalizedPixelPercentages.Contains(faceDistanceEncoding.NormalizedPixelPercentage.Value))
                    continue;
            }
            if (_IdThenNormalizedPixelPercentageToPersonContainers.TryGetValue(faceDistanceEncoding.Id, out normalizedPixelPercentageToPersonContainers))
            {
                if (normalizedPixelPercentageToPersonContainers.ContainsKey(faceDistanceEncoding.NormalizedPixelPercentage.Value))
                    continue;
            }
            faceDistanceLength = faceDistanceLengths[j];
            if (faceDistanceLength.NormalizedPixelPercentage is null || faceDistanceLength.Length is null)
                throw new NotSupportedException();
            if (faceDistanceLength.Length == 0)
                continue;
            if (!_IdThenNormalizedPixelPercentageToPersonContainers.TryGetValue(faceDistanceLength.Id, out normalizedPixelPercentageToPersonContainers))
                continue;
            if (!normalizedPixelPercentageToPersonContainers.ContainsKey(faceDistanceLength.NormalizedPixelPercentage.Value))
                continue;
            personKeysRangesCollection = GetPersonKeysRangesCollection(normalizedPixelPercentageToPersonContainers[faceDistanceLength.NormalizedPixelPercentage.Value]);
            sorting = ISorting.Get(_Configuration.FaceDistancePermyriad, _Configuration.FaceDistanceTolerance, faceDistanceEncoding, faceDistanceLength, personKeysRangesCollection);
            if (sorting.DistancePermyriad == 0)
                continue;
            if (sorting.Id == faceDistanceEncoding.Id)
            {
                if (sorting.NormalizedPixelPercentage == faceDistanceEncoding.NormalizedPixelPercentage.Value)
                    throw new NotSupportedException();
                continue;
            }
            results.Add(sorting);
        }
        return results;
    }

    private int UpdateFromSortingContainers(SortingContainer[] sortingContainers)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        int result = 0;
        string key;
        const int zero = 0;
        string mappingSegmentB;
        string personKeyFormatted;
        PersonBirthday personBirthday;
        PersonContainer[] personContainers;
        const int by = Stateless.IMapLogic.Sorting;
        List<int> normalizedPixelPercentageCollection;
        Dictionary<string, int> checkKeyValuePairs = new();
        Dictionary<int, List<int>> idToNormalizedPixelPercentageCollection = new();
        Dictionary<int, PersonContainer[]>? normalizedPixelPercentageToPersonContainers;
        int totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - _Ticks).TotalSeconds);
        string message = $") {sortingContainers.Length:000} Update From Sorting Container(s) - {totalSeconds} total second(s)";
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        using ProgressBar progressBar = new(sortingContainers.Length, message, options);
        foreach (SortingContainer sortingContainer in sortingContainers)
        {
            progressBar.Tick();
            if (sortingContainer.Mapping is null)
                throw new NotSupportedException();
            if (!_IdThenNormalizedPixelPercentageToPersonContainers.TryGetValue(sortingContainer.Sorting.Id, out normalizedPixelPercentageToPersonContainers))
                throw new NotSupportedException();
            if (!normalizedPixelPercentageToPersonContainers.ContainsKey(sortingContainer.Sorting.NormalizedPixelPercentage))
                throw new NotSupportedException();
            if (!idToNormalizedPixelPercentageCollection.ContainsKey(sortingContainer.Mapping.MappingFromItem.Id))
                idToNormalizedPixelPercentageCollection.Add(sortingContainer.Mapping.MappingFromItem.Id, new());
            normalizedPixelPercentageCollection = idToNormalizedPixelPercentageCollection[sortingContainer.Mapping.MappingFromItem.Id];
            if (normalizedPixelPercentageCollection.Contains(sortingContainer.Mapping.MappingFromLocation.NormalizedPixelPercentage))
                continue;
            personContainers = normalizedPixelPercentageToPersonContainers[sortingContainer.Sorting.NormalizedPixelPercentage];
            foreach (PersonContainer personContainer in personContainers)
            {
                if (personContainer.Key is null || personContainer.Birthdays is null || !personContainer.Birthdays.Any())
                    continue;
                personBirthday = personContainer.Birthdays[zero];
                personKeyFormatted = IPersonBirthday.GetFormatted(_Configuration.PersonBirthdayFormat, personBirthday);
                mappingSegmentB = Stateless.MapLogic.GetMappingSegmentB(_Ticks, personBirthday, personContainer.ApproximateYears, sortingContainer.Mapping.MappingFromItem);
                key = string.Concat(personKeyFormatted, '\t', mappingSegmentB);
                if (!checkKeyValuePairs.ContainsKey(key))
                    checkKeyValuePairs.Add(key, new());
                checkKeyValuePairs[key]++;
                if (checkKeyValuePairs[key] > _Configuration.SortingMaximumPerKey)
                    continue;
                normalizedPixelPercentageCollection.Add(sortingContainer.Mapping.MappingFromLocation.NormalizedPixelPercentage);
                sortingContainer.Mapping.UpdateMappingFromPerson(personContainer.ApproximateYears, by, personContainer.DisplayDirectoryName, personBirthday, mappingSegmentB);
                sortingContainer.Mapping.SetSortingContainer(sortingContainer);
                result += 1;
                break;
            }
        }
        return result;
    }

    internal void ForceSingleImage(Mapping[] mappingCollection)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        const int zero = 0;
        string mappingSegmentB;
        PersonBirthday personBirthday;
        List<int>? normalizedPixelPercentages;
        int by = Stateless.IMapLogic.ForceSingleImage;
        string displayDirectoryName = _Configuration.MappingDefaultName;
        Mapping[] mappingCollectionOrderedConfidence = (from l in mappingCollection orderby l.MappingFromLocation.Confidence descending select l).ToArray();
        foreach (Mapping mapping in mappingCollectionOrderedConfidence)
        {
            if (mapping.MappingFromPerson.PersonBirthday is not null)
                continue;
            if (_SkipCollection.TryGetValue(mapping.MappingFromItem.Id, out normalizedPixelPercentages))
            {
                if (normalizedPixelPercentages.Contains(mapping.MappingFromLocation.NormalizedPixelPercentage))
                    continue;
            }
            foreach (PersonContainer personContainer in _NotMappedPersonContainers)
            {
                if (personContainer.Key is null || personContainer.Birthdays is null || !personContainer.Birthdays.Any())
                    continue;
                personBirthday = personContainer.Birthdays[zero];
                mappingSegmentB = Stateless.MapLogic.GetMappingSegmentB(_Ticks, personBirthday, personContainer.ApproximateYears, mapping.MappingFromItem);
                mapping.UpdateMappingFromPerson(personContainer.ApproximateYears, by, personContainer.DisplayDirectoryName, personBirthday, mappingSegmentB);
                break;
            }
            break;
        }
    }

    public void ForceSingleImageThenSaveSorting(string dFacesContentDirectory, string d2FacePartsContentDirectory, Mapping[] mappingCollection, Dictionary<int, Dictionary<int, Mapping>> idToNormalizedPixelPercentageToMapping, SortingContainer[] sortingContainers, int? useFiltersCounter, int totalNotMapped)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        int updated;
        List<SaveContainer> saveContainers;
        if (!sortingContainers.Any())
        {
            updated = 0;
            ForceSingleImage(mappingCollection);
            saveContainers = Stateless.MapLogic.GetSaveContainers(_Configuration, _EDistanceContentTicksDirectory, dFacesContentDirectory, d2FacePartsContentDirectory, mappingCollection, idToNormalizedPixelPercentageToMapping, useFiltersCounter, saveMapped: false);
        }
        else
        {
            updated = UpdateFromSortingContainers(sortingContainers);
            if (useFiltersCounter is null && totalNotMapped - updated > 0)
                ForceSingleImage(mappingCollection);
            saveContainers = Stateless.MapLogic.GetSaveContainers(_Configuration, _EDistanceContentTicksDirectory, dFacesContentDirectory, d2FacePartsContentDirectory, mappingCollection, idToNormalizedPixelPercentageToMapping, useFiltersCounter, saveMapped: false);
        }
        SaveContainers(totalNotMapped, updated, saveContainers);
    }

    public void CopyManualFiles(string dFacesContentDirectory, Dictionary<int, Dictionary<int, Mapping>> idToNormalizedPixelPercentageToMapping)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        int? id;
        Mapping mapping;
        string faceFile;
        string checkFile;
        string directory;
        FileInfo fileInfo;
        const int zero = 0;
        string faceFileName;
        string shortcutFile;
        string? directoryName;
        string facesDirectory;
        string mappingSegmentB;
        string personDirectory;
        string personKeyFormatted;
        string personDisplayFileName;
        PersonBirthday personBirthday;
        string? personDisplayDirectory;
        int? normalizedPixelPercentage;
        WindowsShortcut windowsShortcut;
        Dictionary<int, PersonContainer[]>? keyValuePairs;
        string by = nameof(Stateless.IMapLogic.ManualCopy);
        Dictionary<int, Mapping>? normalizedPixelPercentageToMapping;
        string successfull = $"_ {nameof(Stateless.IMapLogic.ManualCopy).Humanize(LetterCasing.Title)} Successfull";
        foreach (KeyValuePair<long, PersonContainer> keyValuePair in _PersonKeyToPersonContainer)
        {
            if (keyValuePair.Value.Key is null || keyValuePair.Value.Birthdays is null || !keyValuePair.Value.Birthdays.Any())
                continue;
            personBirthday = keyValuePair.Value.Birthdays[zero];
            foreach (string personDisplayDirectoryAllFile in keyValuePair.Value.DisplayDirectoryAllFiles)
            {
                if (!personDisplayDirectoryAllFile.EndsWith(_Configuration.FacesFileNameExtension))
                    continue;
                (id, normalizedPixelPercentage, _) = IMapping.GetReversedDeterministicHashCodeKey(_Configuration.FacesFileNameExtension, personDisplayDirectoryAllFile);
                if (id is null || normalizedPixelPercentage is null)
                    continue;
                if (_IdThenNormalizedPixelPercentageToPersonContainers.TryGetValue(id.Value, out keyValuePairs))
                {
                    if (keyValuePairs.ContainsKey(normalizedPixelPercentage.Value))
                        continue;
                }
                fileInfo = new(personDisplayDirectoryAllFile);
                if (!fileInfo.Exists)
                    continue;
                personDisplayFileName = Path.GetFileName(personDisplayDirectoryAllFile);
                personDisplayDirectory = Path.GetDirectoryName(personDisplayDirectoryAllFile);
                personKeyFormatted = IPersonBirthday.GetFormatted(_Configuration.PersonBirthdayFormat, personBirthday);
                mappingSegmentB = Stateless.MapLogic.GetMappingSegmentB(_Ticks, personBirthday, keyValuePair.Value.ApproximateYears, fileInfo.CreationTime, isWrongYear: null);
                directory = Path.Combine(_EDistanceContentTicksDirectory, by, personKeyFormatted, mappingSegmentB);
                personDirectory = Path.Combine(directory, keyValuePair.Value.DisplayDirectoryName, "lnk");
                if (!idToNormalizedPixelPercentageToMapping.TryGetValue(id.Value, out normalizedPixelPercentageToMapping))
                    continue;
                if (!normalizedPixelPercentageToMapping.ContainsKey(normalizedPixelPercentage.Value))
                    continue;
                mapping = normalizedPixelPercentageToMapping[normalizedPixelPercentage.Value];
                if (string.IsNullOrEmpty(personDisplayDirectory))
                    throw new NotSupportedException();
                directoryName = Path.GetDirectoryName(mapping.MappingFromItem.RelativePath);
                if (string.IsNullOrEmpty(directoryName))
                    throw new NotSupportedException();
                shortcutFile = Path.Combine(personDisplayDirectory, $"{personDisplayFileName}.lnk");
                facesDirectory = Path.Combine($"{dFacesContentDirectory}{directoryName}", mapping.MappingFromItem.ImageFileHolder.NameWithoutExtension);
                faceFileName = $"{mapping.MappingFromLocation.DeterministicHashCodeKey}{mapping.MappingFromItem.ImageFileHolder.ExtensionLowered}{_Configuration.FacesFileNameExtension}";
                checkFile = Path.Combine(directory, fileInfo.Name);
                if (!Directory.Exists(personDirectory))
                    _ = Directory.CreateDirectory(personDirectory);
                if (!File.Exists(checkFile))
                    File.Copy(personDisplayDirectoryAllFile, checkFile);
                if (personDisplayDirectoryAllFile.Contains(successfull))
                    continue;
                directoryName = Path.Combine(personDisplayDirectory, successfull);
                if (!Directory.Exists(directoryName))
                    _ = Directory.CreateDirectory(directoryName);
                checkFile = Path.Combine(directoryName, personDisplayFileName);
                File.Move(personDisplayDirectoryAllFile, checkFile);
                faceFile = Path.Combine(facesDirectory, faceFileName);
                if (!File.Exists(faceFile))
                    continue;
                if (File.Exists(shortcutFile))
                    continue;
                windowsShortcut = new() { Path = faceFile };
                windowsShortcut.Save(shortcutFile);
                windowsShortcut.Dispose();
                if (!File.Exists(shortcutFile))
                    continue;
                File.SetLastWriteTime(shortcutFile, mapping.MappingFromItem.MinimumDateTime);
            }
        }
    }

    public void SaveNotMappedTicks()
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        int updated = 0;
        string directory;
        const int zero = 0;
        int totalNotMapped = 0;
        string personKeyFormatted;
        SaveContainer saveContainer;
        PersonBirthday personBirthday;
        List<SaveContainer> saveContainers = new();
        const string facePopulatedKey = nameof(Stateless.IMapLogic.Sorting);
        foreach (PersonContainer personContainer in _NotMappedPersonContainers)
        {
            if (personContainer.Key is null || personContainer.Birthdays is null || !personContainer.Birthdays.Any())
                continue;
            personBirthday = personContainer.Birthdays[zero];
            personKeyFormatted = IPersonBirthday.GetFormatted(_Configuration.PersonBirthdayFormat, personBirthday);
            directory = Path.Combine(_EDistanceContentTicksDirectory, $"{facePopulatedKey}NotMapped", personKeyFormatted, _Configuration.MappingDefaultName);
            saveContainer = new(directory);
            saveContainers.Add(saveContainer);
        }
        SaveContainers(totalNotMapped, updated, saveContainers);
    }

    private List<(Mapping, List<(string?, string, string?, string?, string)>)> GetCollection(Mapping[] mappingCollection)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        List<(Mapping, List<(string?, string, string?, string?, string)>)> results = new();
        int years;
        string directory;
        TimeSpan? timeSpan;
        string? copyFileName;
        string? relativePath;
        string? copyDirectory;
        string isWrongYearFlag;
        string shortcutFileName;
        string subDirectoryName;
        string personKeyFormatted;
        DateTime dateTime = DateTime.Now;
        PersonContainer[] personContainers;
        Dictionary<int, PersonContainer[]>? idToPersonContainers;
        List<(string?, string, string?, string?, string)> collection;
        foreach (Mapping mapping in mappingCollection)
        {
            collection = new();
            copyFileName = null;
            copyDirectory = null;
            relativePath = Path.GetDirectoryName($"C:{mapping.MappingFromItem.RelativePath}");
            if (string.IsNullOrEmpty(relativePath) || relativePath.Length < 3)
                continue;
            if (!_IdThenNormalizedPixelPercentageToPersonContainers.TryGetValue(mapping.MappingFromItem.Id, out idToPersonContainers))
            {
                directory = Path.Combine(_EDistanceContentTicksDirectory, $"Unnamed{relativePath[2..]}");
                shortcutFileName = Path.Combine(directory, $"{mapping.MappingFromItem.Id}.lnk");
                collection.Add(new(null, directory, copyDirectory, copyFileName, shortcutFileName));
            }
            else
            {
                if (!idToPersonContainers.ContainsKey(mapping.MappingFromLocation.NormalizedPixelPercentage))
                    continue;
                isWrongYearFlag = IItem.GetWrongYearFlag(mapping.MappingFromItem.IsWrongYear);
                subDirectoryName = $"{isWrongYearFlag}{mapping.MappingFromItem.MinimumDateTime:yyyy}";
                personContainers = idToPersonContainers[mapping.MappingFromLocation.NormalizedPixelPercentage];
                foreach (PersonContainer personContainer in personContainers)
                {
                    if (personContainers.Length != 1)
                        break;
                    if (personContainer.Key is null || personContainer.Birthdays is null || !personContainer.Birthdays.Any())
                        continue;
                    foreach (PersonBirthday personBirthday in personContainer.Birthdays)
                    {
                        timeSpan = IPersonBirthday.GetTimeSpan(mapping.MappingFromItem.MinimumDateTime, mapping.MappingFromItem.IsWrongYear, personBirthday);
                        if (timeSpan.HasValue)
                        {
                            if (timeSpan.Value.Ticks < 0)
                                subDirectoryName = "!---";
                            else
                            {
                                (years, _) = IPersonBirthday.GetAge(dateTime, personBirthday);
                                subDirectoryName = $"^{years:000}";
                            }
                        }
                        personKeyFormatted = IPersonBirthday.GetFormatted(_Configuration.PersonBirthdayFormat, personBirthday);
                        directory = Path.Combine(_EDistanceContentTicksDirectory, "Shortcuts", personKeyFormatted, subDirectoryName);
                        copyDirectory = Path.Combine(_EDistanceContentTicksDirectory, "Images", personKeyFormatted, subDirectoryName);
                        copyFileName = Path.Combine(copyDirectory, $"{mapping.MappingFromItem.Id}{mapping.MappingFromItem.ResizedFileHolder.ExtensionLowered}");
                        shortcutFileName = Path.Combine(directory, $"{mapping.MappingFromItem.Id}.lnk");
                        collection.Add(new(personKeyFormatted, directory, copyDirectory, copyFileName, shortcutFileName));
                    }
                }
                if (!collection.Any())
                {
                    copyFileName = null;
                    copyDirectory = null;
                    directory = Path.Combine(_EDistanceContentTicksDirectory, $"Not Supported{relativePath[2..]}", subDirectoryName);
                    shortcutFileName = Path.Combine(directory, $"{mapping.MappingFromItem.Id}.lnk");
                    collection.Add(new(null, directory, copyDirectory, copyFileName, shortcutFileName));
                }
            }
            results.Add(new(mapping, collection));
        }
        return results;
    }

    public void SaveShortcuts(string[] juliePhares, Mapping[] mappingCollection)
    {
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        List<(Mapping, List<(string?, string, string?, string?, string)>)> facesToCollection = GetCollection(mappingCollection);
        foreach ((Mapping mapping, List<(string? PersonKeyFormatted, string, string?, string?, string)> collection) in facesToCollection)
        {
            if (collection.Any(l => l.PersonKeyFormatted is not null and "1501-04-10_00"))
                continue;
            foreach ((string? personKeyFormatted, string directory, string? copyDirectory, string? copyFileName, string _) in collection)
            {
                if (personKeyFormatted is null)
                    continue;
                if (!Directory.Exists(directory))
                    _ = Directory.CreateDirectory(directory);
                if (string.IsNullOrEmpty(copyDirectory) || string.IsNullOrEmpty(copyFileName))
                    continue;
                if (!juliePhares.Contains(personKeyFormatted))
                    continue;
                if (!Directory.Exists(copyDirectory))
                    _ = Directory.CreateDirectory(copyDirectory);
                if (!File.Exists(copyFileName))
                    File.Copy(mapping.MappingFromItem.ResizedFileHolder.FullName, copyFileName);
                break;
            }
        }
    }

}